/*
 * Implement a null-terminated string buffer.
 *
 * Destructor don't deallocate string buffer because is considered
 * the case where this buffer is shared with others.
 * 
 * FEATURES:
 * 
 * automaticaly manage dynamic allocated memory for string
 *
 * LICENSE:
 *
 * Copyright (C) 2000 Leandro Bernsmller
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 * 
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */



#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include "../../include/base/type.h"
#include "../../include/base/string.h"



/*
 * constructor - init class variables
 */
class_string::class_string ()
{
	init ();
}



/*
 * destructor
 */
class_string::~class_string ()
{
}



/*
 * set default values
 */
void  class_string::
init ()
{
	m_data = 00;
	m_alloc_size = 0;
}



/*
 * Clear, empty, the string.
 */
void  class_string::
terminate ()
{
	if (m_data) {
		free (m_data);
		m_data = 00;
		m_alloc_size = 0;
	}
}



/*
 * Resize buffer if it isn't big enough to grow fitting more the
 * given number of bytes.
 * more - number of bytes to test
 * return:
 * TRUE - ok or resize not needed /
 * FALSE - error: not enough system memory to make resize
 * (let untouched the current buffer)
 */
inline BOOL class_string::
size_buffer (DWORD more)
{
	// buffer is free
	if (!m_data) {

		m_data = (char *) malloc (more + 1);

		if (!m_data)
			return FALSE;

		*m_data = 0;
		m_alloc_size = more + 1;

		return TRUE;
	}

	// resize buffer
	DWORD used = strlen (m_data);
	if (used + more + 1 >= m_alloc_size) {

		DWORD step_size = (more / ALLOC_STEP + 1) * ALLOC_STEP;

		char *new_buffer = (char *) realloc (m_data, used + step_size);

		if (!new_buffer)
			return FALSE;

		m_data = new_buffer;
		m_alloc_size = used + step_size;
	}

	return TRUE;
}



/*
 * Get buffer contents.
 * return:
 * the string, or NULL if not allocated 
 */
char *class_string::
get ()
{
	return m_data;
}



/*
 * Set buffer contents with a copy of source string.
 * return:
 * buffer contents or 00 if can't copy, not enough memory
 */
char*  class_string::
copy (const char *str)
{
	DWORD str_size = strlen (str);
	if (str_size + 1 >= m_alloc_size) {

		BOOL  result = size_buffer (str_size + 1 - m_alloc_size);

		if (!result)
			return 00;
	}

	strcpy (m_data, str);
	return m_data;
}



/*
 * Duplicates the given string: makes one copy, and set buffer
 * contents with this copy.
 * return:
 * buffer contents or 00 if can't copy, not enough memory
 */
char*  class_string::
duplicate (const char *str)
{
  init ();

  DWORD str_size = strlen (str);
  if (str_size + 1 >= m_alloc_size) {

    BOOL  result = size_buffer (str_size + 1 - m_alloc_size);

    if (!result)
      return 00;
  }

  strcpy (m_data, str);
  return m_data;
}


/*
 * Resize buffer freeing unused memory.
 */
void  class_string::
free_unused ()
{
	m_data = (char *) realloc (m_data, strlen (m_data) + 1);
}



/*
 * Append the given string to end.
 * return:
 * TRUE - ok / FALSE - can't append, not enough memory
 */
BOOL  class_string::
append (const char *str)
{
	BOOL  result = size_buffer (strlen (str));

	if (!result)
		return FALSE;

	strcat (m_data, str);

	return TRUE;
}



/*
 * Include one character.
 * return:
 * TRUE - ok / FALSE - can't include, not enough memory
 */
BOOL  class_string::
append_char (char character)
{
	BOOL  result = size_buffer (1);

	if (!result)
		return FALSE;

	DWORD used = strlen (m_data);

	m_data [used] = character;
	m_data [used + 1] = 0;

	return TRUE;
}



/* Append the given string to buffer end. Replace
   printf string output format (%s) in first string with
   the second given string.

  str - the string where replace will happen
  form - the string to replace

  return:
  TRUE - ok / FALSE - error
*/
BOOL  class_string::
append_form (char* str, char* form)
{
  DWORD grow_size;
  grow_size = strlen (str) + strlen (form) - 2;  // -2 because %s will go out

  BOOL result = size_buffer (grow_size);
 
  if (!result)
    return FALSE;

  // define from where start formating
  char* format_pos = (char*) m_data + strlen (m_data);

  // append
  strcat (m_data, str);

  // format (replace)
  sprintf (format_pos, "%s", form);

  return TRUE;
}
