/*
 * The lists component parser.
 * Handle the recognition of all types of lists during script parsing.
 * 
 * LICENSE:
 *
 * Copyright (C) 2000 Leandro Bernsmller
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 * 
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */



#include <stdio.h>
#include "../../include/sbase.h"
#include "../../include/lao/laodefs.h"
#include "../../include/lao/colist.h"
#include "../../include/lao/colist_l.h"
#include "../../include/lao/colist_i.h"
#include "../../include/lao/colist_e.h"
#include "../../include/lao/colist_d.h"
#include "../../include/lao/plist.h"



class_plist::class_plist ()
{
  init ();
}



class_plist::~class_plist ()
{
  terminate ();
}



void  class_plist::
init ()
{
  m_state = 0;
  m_list_type = LAOL_UNDEFINED;
  m_inner_type = LAOL_UNDEFINED;
}



void  class_plist::
terminate ()
{
  init ();
}



/*
 * Return the list type of the list being parsed.
 * return:
 * list type
 */
DATAPT class_plist::
get_list_type ()
{
  return (DATAPT) m_list_type;
}



/*
 * Set the list type of the list being parsed.
 */
void  class_plist::
set_list_type (DATAPT lt)
{
  m_list_type = (l_list_type) lt;
}



/*
 * Return the list type of the inner list being parsed.
 * return:
 * list type
 */
DATAPT class_plist::
get_inner_type ()
{
  return (DATAPT) m_inner_type;
}



/*
 * Set the type of a list which begin is inside this list.
 */
void  class_plist::
set_inner_type (DATAPT it)
{
  m_inner_type = (l_list_type) it;
}



/*
 * Test if end of list was found, if yes update state machine.
 * sym - symbol read
 */
DATAPT  class_plist::
process_list_end (symbol_info* sym)
{
  class_char* cho = &(sym->obj);
  char ch = cho->get ();
  l_sysmsg ret_val = NONE;

  if (sym->ct->match (ch, CHC_CLOSE_BRACKET) ||      // end list
      sym->ct->match (ch, CHC_CLOSE_PARENTHESIS) ||
      sym->ct->match (ch, CHC_CLOSE_BRACE) ||
      sym->ct->match (ch, CHC_CLOSE_RELATIONAL)) {

    ret_val = set_msgsrv (BIN_RECOGNITION_TOKEN_SUCC, 3);
    m_state = 2;

    if (sym->ct->match (ch, CHC_CLOSE_BRACKET) &&
        get_list_type () != LAOL_IMMEDIATE) {
      ret_val = BIN_RECOGNITION_FAIL;
    }

    if (sym->ct->match (ch, CHC_CLOSE_BRACE) &&
        get_list_type () != LAOL_DIRECT) {
      ret_val = BIN_RECOGNITION_FAIL;
    }

    if (sym->ct->match (ch, CHC_CLOSE_PARENTHESIS) &&
        get_list_type () != LAOL_INDIRECT) {
      ret_val = BIN_RECOGNITION_FAIL;
    }

    if (sym->ct->match (ch, CHC_CLOSE_RELATIONAL) &&
        get_list_type () != LAOL_LITERAL) {
      ret_val = BIN_RECOGNITION_FAIL;
    }
  }

  return (DATAPT) ret_val;
}



/*
 * Makes the state machine process the given symbol.
 * sym - symbol to process
 * return:
 * machine result
 */
DATAPT  class_plist::
process_symbol (symbol_info* sym)
{
  class_char* cho = &(sym->obj);
  char ch = cho->get ();
  l_sysmsg ret_val = NONE;
 
  switch (m_state) {
  case 0:   // start or end list recognition
    {
      if (! sym->ct->match (ch, CHC_SEPARATORS)) { // if an invalid char was read
        ret_val = BIN_RECOGNITION_FAIL;
        m_state = 2;
      }

      if (sym->ct->match (ch, CHC_OPEN_BRACKET)) {
        ret_val = set_msgsrv (BIN_RECOGNITION_TOKEN_SUCC, 2);
        set_inner_type (LAOL_IMMEDIATE);
        m_state = 1;
      }

      if (sym->ct->match (ch, CHC_OPEN_BRACE)) {
        ret_val = set_msgsrv (BIN_RECOGNITION_TOKEN_SUCC, 2);
        set_inner_type (LAOL_DIRECT);
        m_state = 1;
      }

      if (sym->ct->match (ch, CHC_OPEN_PARENTHESIS)) {
        ret_val = set_msgsrv (BIN_RECOGNITION_TOKEN_SUCC, 2);
        set_inner_type (LAOL_INDIRECT);
        m_state = 1;
      }

      if (sym->ct->match (ch, CHC_OPEN_RELATIONAL)) {
        ret_val = set_msgsrv (BIN_RECOGNITION_TOKEN_SUCC, 2);
        set_inner_type (LAOL_LITERAL);
        m_state = 1;
      }

      DATAPT ret_end = process_list_end (sym);
      if (ret_end != NONE) ret_val = (l_sysmsg) ret_end;
    }

    // wait next input
    break;

  case 1:         // final, beginning from start structure recognition
    break;

  case 2:         // final, beginning from end structure recognition
    break;
  }

  return (DATAPT) ret_val;
}



/*
 * Return the component which represents the recognized
 * token or structure.
 * return:
 * pointer to the returned component or data
 */
DATAPT  class_plist::
return_component ( class_enviroment* env, class_sbase_history* hist )
{
  class_component* comp = 00;
  if (m_state == 1) {  // structure begin recognized

    // create and return one specific list component
    switch (get_inner_type ()) {

    case LAOL_IMMEDIATE:
      {
        class_colist* lst = (class_colist*) new (class_colist_immediate);
        comp = ::build_component ((DATAPT) lst, env, hist,
          ::call_list,
          "plist", "list_immediate");
      }
      break;

    case LAOL_DIRECT:
      {
        class_colist* lst = (class_colist*) new (class_colist_direct);
        comp = ::build_component ((DATAPT) lst, env, hist,
          ::call_list,
          "plist", "list_direct");
      }
      break;

    case LAOL_INDIRECT:
      {
        class_colist* lst = (class_colist*) new (class_colist_indirect);
        comp = ::build_component ((DATAPT) lst, env, hist,
        ::call_list,
        "plist", "list_literal");
      }
      break;

    case LAOL_LITERAL:
      {
        class_colist* lst = (class_colist*) new (class_colist_literal);
        comp = ::build_component ((DATAPT) lst, env, hist,
          ::call_list,
          "plist", "list_literal");
      }
      break;

    case LAOL_UNDEFINED:
      comp = 00; // REVIEW - error, try return one list without type
    }
  }

  return (DATAPT) comp;
}



/*
 * Reset the parse state machine to the initial state.
 * Preparing to begin a token recognition.
 * env - enviroment
 * hist - history
 * return:
 * NULL
 */
DATAPT  class_plist::
sync_recognition ( class_enviroment* env, class_sbase_history* hist )
{
  m_state = 0;
  return 00;
}



/*
 * When a parser signals success in recognition,
 * is tested with this mean success in end of the
 * structure, too.
 * sp - component (parser) which signals success
 * env - enviroment
 * hist - history
 * return:
 * recognition result
 */
DATAPT  class_plist::
recognition_token_succ ( class_component* sp,
                         class_enviroment* env,
                         class_sbase_history* hist )
{
  // verify if the given parser is for string type
  if (strcmp ( "pstring", sp->get_self_name () ) == 0) {

    // verify if string success was achieved reading any close structure

    // get the read symbol information
    DATAPT msg = set_msgsrv (BIN_ACTIVATE, 1);
    symbol_info* sym = (symbol_info*) sp->call (msg, env, hist, 00);

    // the structure state machine must change: close structure found
    return process_list_end (sym);
  }

  return NONE;
}



/*
 * Function to handle component messages.
 * self - class_component instance for this object
 * msg - lao system message ID
 * env - reference to lao name table
 * hist - reference to history
 * pdata - points component data
 * param - component parameter
 *
 * return:
 * process result
 */
DATAPT
call_plist (class_component* self, DATAPT msg, class_enviroment* env,
            class_sbase_history* hist, DATAPT* pdata, DATAPT param)
{

  DATA_POINTED (pdata, hist);

  switch (get_sysmsg (msg)) {

  case BIN_CREATE_DATA:
    {
      DATA_UNAVAILABLE (*pdata, hist);

      *pdata = (DATAPT) malloc (sizeof (class_plist));
      MEMORY_ALLOCATED (*pdata, hist);

      // define information
      class_plist *p = (class_plist *) * pdata;
      p->init ();
      p->set_list_type (param);

      return *pdata;
    }
    break;

  case BIN_CLEAN_DATA:
    {
      DATA_AVAILABLE (*pdata, hist);

      class_plist *p = (class_plist *) * pdata;
      p->terminate ();

      free ((void *) *pdata);
      *pdata = 0;
      return *pdata;
    }
    break;

  case BIN_PROCESS_SYMBOL:
    {
      DATA_AVAILABLE (*pdata, hist);

      symbol_info* sym = (symbol_info*) param;
  
      class_plist *p = (class_plist *) * pdata;
      return p->process_symbol (sym);
    }
    break;    

  case BIN_RETURN_COMPONENT:
    {
      DATA_AVAILABLE (*pdata, hist);

      class_plist *olst = (class_plist *) * pdata;
      return olst->return_component (env, hist);
    }
    break;    

  case BIN_SYNC_RECOGNITION:
    {
      DATA_AVAILABLE (*pdata, hist);

      class_plist *olst = (class_plist *) * pdata;
      return olst->sync_recognition (env, hist);
    }
    break;

  case BIN_RECOGNITION_TOKEN_SUCC:
    {
      DATA_AVAILABLE (*pdata, hist);

      if (param) {

        class_plist* olst = (class_plist *) * pdata;
        class_component* comp = (class_component*) param;
        return olst->recognition_token_succ (comp, env, hist);
      }
    }
    break;
      
  case BIN_RECOGNITION_TOKEN_BEGIN:
  case BIN_RECOGNITION_FAIL:
  case BIN_ADD_COMPONENT:
  case BIN_ACTIVATE:
  case INVOKE:
  case ACTIVATE:
  case ACTIVATE_AS_REFERENCE:
  case ACTIVATE_AS_PARAMETER:
    {
       return 00;
    }
    break;

  case BIN_GET_TEXT_REPRESENTATION:
    {
      class_string *str = (class_string *) malloc (sizeof (class_string));
      str->init ();
      str->append ("class_plist instance");
      return (DATAPT) str;
    }
    break;

  case COPY:
    {
      class_plist *olst = (class_plist *) * pdata;
      return (DATAPT) ::build_component ((DATAPT) olst->get_inner_type (), env, hist,
                                ::call_plist,
                                "plist", "plist");
    }
    break;
  }

  return 00;
}
