/*
 * Implements a string parser as a component.
 * Handle strings recognition during script parsing.
 * 
 * LICENSE:
 *
 * Copyright (C) 2000 Leandro Bernsmller
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 * 
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */



#include <stdio.h>
#include "../../include/sbase.h"
#include "../../include/lao/laodefs.h"
#include "../../include/lao/costring.h"
#include "../../include/lao/pstring.h"



class_pstring::class_pstring ()
{
  init ();
}



class_pstring::~class_pstring ()
{
  terminate ();
}



void  class_pstring::
init ()
{
  m_str.init ();
  m_state = 0;
  m_sym = 00;
}



void  class_pstring::
terminate ()
{
  m_str.terminate ();
  init ();
}



/*
 * Makes the state machine process the given symbol.
 * sym - symbol to process
 * return:
 * machine result
 */
DATAPT  class_pstring::
process_symbol (symbol_info* sym)
{
  m_sym = sym;
  class_char* cho = &(sym->obj);
  char ch = cho->get ();
  l_sysmsg ret_val = NONE;

  switch (m_state) {

  case 0:   // starting string recognition
    if (sym->ct->match (ch, CHC_STRING_DELIMITER)) { // begin
      ret_val = BIN_RECOGNITION_TOKEN_BEGIN;
      m_state = 1;
      m_str.terminate ();
    } else {
      if (sym->ct->match (ch, CHC_INVALID_STRING_START)) {  // end, fail
        ret_val = BIN_RECOGNITION_FAIL; 
        m_state = 2;
      } else {
        if (! sym->ct->match (ch, CHC_SEPARATORS)) {  // starting string without ""
          ret_val = BIN_RECOGNITION_TOKEN_BEGIN;
          m_state = 3;
          m_str.terminate ();
          m_str.append_char (ch);
        }
      }
    }
    // wait next input, if character is a separator
    break;

    case 1:   // process string started with a delimiter (")
      if (sym->ct->match (ch, CHC_STRING_DELIMITER)) {  // end, success
        ret_val = BIN_RECOGNITION_TOKEN_SUCC;
        m_state = 2;
      } else {            // continue
        m_str.append_char (ch);
      }
      break;

    case 2:   // string recognition ended
      break;

    case 3:   // string started by "OTHERS" chars (letters,...)
      if (sym->ct->match (ch, CHC_SEPARATORS)        ||
          sym->ct->match (ch, CHC_CLOSE_PARENTHESIS) ||
          sym->ct->match (ch, CHC_CLOSE_BRACE)       ||
          sym->ct->match (ch, CHC_CLOSE_BRACKET)     ||
          sym->ct->match (ch, CHC_CLOSE_RELATIONAL) ) {  // end, success
        ret_val = BIN_RECOGNITION_TOKEN_SUCC;
        m_state = 2;
      } else {            // continue
        m_str.append_char (ch);
      }
      break;
  }

  return (DATAPT) ret_val;
}



/*
 * Return the component which represents the recognized
 * token or structure.
 * return:
 * pointer to the returned component or data
 */
DATAPT  class_pstring::
return_component ( class_enviroment* env, class_sbase_history* hist )
{
  return (DATAPT) ::build_component ((DATAPT) m_str.get (), env, hist, ::call_string, "<pstring>", "<string>");
}



/*
 * Reset the parse state machine to the initial state.
 * Preparing to begin a token recognition.
 * return:
 * NULL
 */
DATAPT  class_pstring::
sync_recognition ( class_enviroment* env, class_sbase_history* hist )
{
  m_state = 0;
  return 00;
}



/*
 * Return information for the last symbol read.
 * return:
 * symbol pointer
 */
symbol_info*  class_pstring::
get_symbol_info ()
{
  return m_sym;
}



/*
 * Function to handle component messages.
 * self - class_component instance for this object
 * msg - lao system message ID
 * env - reference to lao name table
 * hist - reference to history
 * pdata - points component data
 * param - component parameter
 *
 * return:
 * process result
 */
DATAPT
call_pstring (class_component * self, DATAPT msg, class_enviroment * env,
              class_sbase_history* hist, DATAPT * pdata, DATAPT param)
{

  DATA_POINTED (pdata, hist);

  switch (get_sysmsg (msg)) {

  case BIN_CREATE_DATA:
    {
        DATA_UNAVAILABLE (*pdata, hist);

        *pdata = (DATAPT) malloc (sizeof (class_pstring));
        MEMORY_ALLOCATED (*pdata, hist);

        // define information
        class_pstring *ostr = (class_pstring *) * pdata;
        ostr->init ();

        return *pdata;
    }
    break;

  case BIN_CLEAN_DATA:
    {
      DATA_AVAILABLE (*pdata, hist);

      class_pstring *ostr = (class_pstring *) * pdata;
      ostr->terminate ();

      free ((void *) *pdata);
      *pdata = 0;
      return *pdata;
   }
    break;

  case BIN_PROCESS_SYMBOL:
    {
      DATA_AVAILABLE (*pdata, hist);

      symbol_info* sym = (symbol_info*) param;
    
      class_pstring *ostr = (class_pstring *) * pdata;
      return (DATAPT) ostr->process_symbol (sym);
    }
  break;    

  case BIN_RETURN_COMPONENT:
    {
      DATA_AVAILABLE (*pdata, hist);

      class_pstring *ostr = (class_pstring *) * pdata;
      return ostr->return_component (env, hist);
    }
  break;    

  case BIN_SYNC_RECOGNITION:
    {
      DATA_AVAILABLE (*pdata, hist);

      class_pstring *ostr = (class_pstring *) * pdata;
      return ostr->sync_recognition (env, hist);
    }
    break;

  case BIN_ACTIVATE:
    {
      DATA_AVAILABLE (*pdata, hist);

      class_pstring *ostr = (class_pstring *) * pdata;

      int res_op = get_actsrv (msg);

      switch (res_op) {
      case 1:     // return information for the last processed symbol
        return (DATAPT) ostr->get_symbol_info ();
        break;
      }
    }
    break;

  case BIN_RECOGNITION_TOKEN_BEGIN:
  case BIN_RECOGNITION_TOKEN_SUCC:
  case BIN_RECOGNITION_FAIL:
  case BIN_ADD_COMPONENT:
  case INVOKE:
  case ACTIVATE:
  case ACTIVATE_AS_REFERENCE:
  case ACTIVATE_AS_PARAMETER:
    {
      return 00;
    }
    break;

  case BIN_GET_TEXT_REPRESENTATION:
    {
      class_string *str = (class_string *) malloc (sizeof (class_string));
      str->init ();
      str->append ("class_pstring instance");
      return (DATAPT) str;
    }
    break;

  case COPY:
    {
      return (DATAPT) build_component ((DATAPT) NULL, env, hist, ::call_pstring, "<pstring>", "<string>");
    }
    break;
  }

  return 00;
}
