/*
 * Base component class. Implement common operations for all components.
 *
 * LICENSE:
 *
 * Copyright (C) 2000 Leandro Bernsmller
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 * 
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */



#include "../../include/base.h"
#include "../../include/sbase/def.h"
#include "../../include/sbase/shistory.h"
#include "../../include/sbase/env.h"
#include "../../include/sbase/comp.h"



#define MSG_CFDEF "call function still defined."
#define MSG_CFNDEF "call function not defined."
#define MSG_DATA_ERROR "component data creation fail."
#define MSG_TRACED "Traced component '%s', parent: '%s', system message id: %d.\n"



/*
 * Allocate and init and return one new component object.
 * This function is designed to be called in C scope,
 * not through a reference to one instanciated object.
 * Example: 
 *   ::build_component ( data, vars, hlog, fn, "foo", "bar");
 * init_data - initial data for component
 * env - enviroment, the name table
 * hist - processing history
 * call_fn - component call function
 * parent_n - component parent name (null terminated string)
 * self_n - component name (null terminated string)
 * return:
 * pointer to the new component or 00 if fail
 */
class_component *
build_component (DATAPT init_data,
         class_enviroment* env,
         class_sbase_history* hist,
         CALLFN (*call_fn),
         char *parent_n,
         char *self_n)
{
    // alloc component
    class_component *comp = (class_component *) malloc (sizeof (class_component));
    MEMORY_ALLOCATED (comp, hist);

    // reset component to initial state
    comp->init ();

    // define component behavior
    BOOL  success = comp->create (parent_n, self_n, call_fn, hist);
    if (success == FALSE) {
      free (comp);
      return 00;
    }

    // set component data
    DATAPT data_success = comp->call (BIN_CREATE_DATA, env, hist, init_data);
    if (!data_success) {
      hist->add_text (LH_TEXT, MSG_DATA_ERROR);
      comp->terminate ();
      free (comp);
      return 00;
    }

    return comp;
}



/*
 * init with empty object
 */
class_component::class_component ()
{
    init ();
}



/*
 * terminate (makes object empty)
 */
class_component::~class_component ()
{
    terminate ();
}



/*
 * default data initialization
 */
void  class_component::
init ()
{
    m_call_function = 0;
    m_data = 0;
    m_parent_name = m_self_name = 0;
}



/*
 * Deallocate object data
 */
void  class_component::
terminate ()
{
  if (m_parent_name)
    free (m_parent_name);

  if (m_self_name)
    free (m_self_name);

  init ();
}



/*
 * Set call function and component names.
 * parent_n - creator name
 * self_n - component name
 * func - the call function
 * hist - sbase history log
 *
 * Names are copyed to memory allocated by component.
 * 
 * return:
 * TRUE - ok / FALSE - can't set call function 
 */
BOOL  class_component::
create (char *parent_n, char *self_n, CALLFN (*func), class_sbase_history* hist)
{
  if (!m_call_function) {

    m_parent_name = (char *) malloc (strlen (parent_n) + 1);
    if (m_parent_name)
      strcpy (m_parent_name, parent_n);
    else {
      hist->add_predef_text (LH_TEXT, "error2");
      hist->add_flag (LH_ERROR, ERR_NO_MEMORY);
      hist->set_exception (LE_EXIT | LE_ERROR | LE_FATAL_ERROR);
      return FALSE;
    }

    m_self_name = (char *) malloc (strlen (self_n) + 1);
    if (m_self_name)
      strcpy (m_self_name, self_n);
    else {
      free (m_parent_name);
      m_parent_name = 00;
      hist->add_predef_text (LH_TEXT, "error2");
      hist->add_flag (LH_ERROR, ERR_NO_MEMORY);
      hist->set_exception (LE_EXIT | LE_ERROR | LE_FATAL_ERROR);
      return FALSE;
    }

    m_call_function = func;

    return TRUE;
  }

  // call function not defined - inform error
  hist->add_text (LH_TEXT, MSG_CFDEF );
  hist->add_flag (LH_ERROR, ERR_UNDEFINED);
  hist->set_exception (LE_EXIT | LE_ERROR);
  return FALSE;
}



/*
 * Try to identify the given string as one operator.
 * Is used to help parameters analisis when one component is
 * activaded.
 * op_str - the given string representation for the operator
 * size - number of operators
 * ops_tab - table (array) with strings and operators IDs.
 * Return operator ID (flag) for the given string or
 * size (normaly the LAST_ID value) if operator was not found.
 */
l_op_id class_component::
parse_param_str (char* op_str, DWORD size, l_op_element* ops_tab)
{
  for ( DWORD index = 0; index < size; index ++ ) {

    l_op_element *op = ops_tab + index;

    if ( strcmp ( op->str, op_str ) == 0 )
      return op->id;
  }

  return size;
}




/*
 * Return the components parent (creator) name.
 */
char* class_component::
get_parent_name ()
{
	return m_parent_name;
}



/*
 * Return the component name (its class type).
 */
char* class_component::
get_self_name ()
{
	return m_self_name;
}



/*
 * Execute the call function.
 * sysmsg - system message flag
 * env - component's enviroment
 * hist - sbase history
 * param - user parameter
 * 
 * return:
 * the call result or 0 (if call function wasn't defined) 
 */
DATAPT class_component::
call (DATAPT sysmsg, class_enviroment * env, class_sbase_history* hist, DATAPT param)
{
  // show debug information
  if (hist->has_debug () )
    if (hist->get_debug_mode () == DB_TRACE)
      printf (MSG_TRACED, m_self_name, m_parent_name, (DWORD) sysmsg );

  if (m_call_function)
    return (*m_call_function) (this, sysmsg, env, hist, &m_data, param);
  
  // call function not defined - error
  hist->add_text (LH_TEXT, MSG_CFNDEF);
  hist->add_flag (LH_ERROR, ERR_UNDEFINED);
  hist->set_exception (LE_EXIT | LE_ERROR);
  return 0;
}
