/*
 * Script translator. Coordenates lexical analise,
 * redirects input to components. These components,
 * then, are responsable to create the binary representation
 * of the given text.
 *
 *
 * LICENSE:
 *
 * Copyright (C) 2000 Leandro Bernsmller
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 * 
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */



#include "../../include/base.h"
#include "../../include/sbase/def.h"
#include "../../include/sbase/shistory.h"
#include "../../include/sbase/env.h"
#include "../../include/sbase/comp.h"
#include "../../include/sbase/mparser.h"
#include "../../include/sbase/strans.h"



class_translator::class_translator ()
{
        init ();
}



class_translator::~class_translator ()
{
}



void  class_translator::
init ()
{
  m_recogs.init ();
  m_structs.init ();
}



/*
 * Terminate translator.
 */
void  class_translator::
terminate (class_enviroment* env, class_sbase_history* hist)
{
  m_recogs.terminate (env, hist);
  m_structs.terminate ();
}



/*
 * Define a parser for recognition of tokens for a specific type.
 * pname - name of the type of tokens to be recognized
 * comp - organizer that recognize the tokens of this type
 * return:
 * TRUE - success; FALSE - error: no memory, or type still defined
 */
BOOL class_translator::
add_type (char* pname, class_component* comp)
{
  return m_recogs.add_type (pname, comp);
}



/*
 * Push one component in stack of structured components.
 */
BOOL class_translator::
push_structure (class_component* comp)
{
  return m_structs.insert_first (BEFORE, get_from comp);
}



/*
 * Pop one component from stack of structured components.
 * return:
 * the component or 00 if none could be poped
 */
class_component* class_translator::
pop_structure ()
{
  class_component* comp;
  BOOL res = m_structs.remove_first (put_in comp);

  if (res)
    return comp;
  else
    return 00;
}



/*
 * Update translator if a token was recognized.
 * par_msg - the message received after symbol processing
 * env - the enviroment
 * hist - execution history
 * ret_comp - the component to return, acording the recognized token
 * end_input - flag if is time to stop input processing
 * end_parse - flag if is time to stop current symbol parsing
 * return:
 * a component, or 00 if none found
 * end_input - TRUE: stop; FALSE: continue
 * end_parse - TRUE: stop; FALSE: continue
 */
class_component* class_translator::
get_processed_token (DATAPT par_msg,
                     class_enviroment* env,
                     class_sbase_history* hist,
                           BOOL* end_input,
                           BOOL* end_parse)
{
  class_component* ret_comp = 00;
  
  // if some parser recognize a token
  if (get_sysmsg (par_msg) == BIN_RECOGNITION_TOKEN_SUCC) {

    // An atomic or a structure-delimiter token were recognized.

    // ask to parser, the recognized component
    ret_comp = (class_component*) m_recogs.activate_type (BIN_RETURN_COMPONENT, env, hist, 00);

    // define the operation to be done with the token found
    int res_op = get_actsrv (par_msg);

    switch (res_op) {
    case 0:   // atomic token recognized
    case 1:   // middle structure delimiter recognized
    case 3:   // close structure delimiter recognized
      {
        // component to send the returned one (it is in stack top)
        class_component* top_comp;
        BOOL res_top = m_structs.get_first (put_in top_comp);

        // try send the new component to the one in stack top
        if (! res_top)
          // stack empty, return the component
          *end_input = TRUE;
        else {

          // define and send message to top component
          DATAPT msg_top = set_msgsrv (BIN_ADD_COMPONENT, res_op);
          BOOL res_st = top_comp->call (msg_top, env, hist, get_from ret_comp);

          // error, component sent was not accepted
          if (! res_st) {
            ret_comp = 00;
            *end_parse = TRUE; // REVIEW history notification?
          }

          // operation: pop stack
          if (res_op == 3) {

            // pop
            class_component* pop_comp = pop_structure ();
            m_recogs.pop_parser ();  // REVIEW trash, garbage collection?

            // if stack is empty, return the poped component
            if (m_structs.get_total_nodes () == 0) {

              *end_parse = TRUE;
              *end_input = TRUE;   // REVIEW: the right place to test?
              ret_comp = pop_comp;

            } else {

              // send to top component the poped one
              res_top = m_structs.get_first (put_in top_comp);
              DATAPT msg_newtop = set_msgsrv (BIN_ADD_COMPONENT, res_op);
              BOOL res_nt = top_comp->call (msg_newtop, env, hist, get_from pop_comp);
              if (! res_nt) {
                ret_comp = 00;
                *end_parse = TRUE;  // REVIEW: history notification
              }
            }
          }
        }
      }
      break;

    case 2:   // start structure delimiter recognized, push a new structure
      push_structure (ret_comp); // REVIEW: history notification about errors
      m_recogs.push_parser (env, hist);
      break;
    }

    *end_parse = TRUE;
  }

  return ret_comp;
}



/*
 * Try parse the given symbol.
 * sym - information about the symbol read
 * env - the enviroment
 * hist - execution history
 * ret_comp - the component to return
 * return:
 * ret_comp - a component, or 00 if none found
 * TRUE - continue parsing; FALSE - stop
 */
BOOL class_translator::
parse_symbol (symbol_info* sym,
              class_enviroment* env,
              class_sbase_history* hist,
              class_component** ret_comp)
{
  BOOL end_symparse = FALSE;
  BOOL end_input = FALSE;
  *ret_comp = 00;

  if (! m_recogs.use_first_type () )  // define first component to send input
    return TRUE;   // no parsers available

  do {  // for each top parser

    // send symbol to component (parser)

    DATAPT res_proc = m_recogs.activate_type (BIN_PROCESS_SYMBOL, env, hist, get_from sym);
    l_sysmsg res_msg = (l_sysmsg) get_sysmsg (res_proc);
    
    // acording the process result
    if (res_msg == BIN_RECOGNITION_TOKEN_BEGIN ||
        res_msg == BIN_RECOGNITION_TOKEN_SUCC  ||
        res_msg == BIN_RECOGNITION_FAIL) {

      // send a message to all top parsers
      class_component* comp = m_recogs.get_parser_in_use ();
      m_recogs.start_activate_all ();
      BOOL end_act_all = FALSE;
      while (! end_act_all) {
        DATAPT res = m_recogs.activate_all ( get_from res_msg, env, hist, get_from comp, TRUE );
        if (res == 00)
          end_act_all = TRUE;
        else
          // process activation result
          *ret_comp = get_processed_token(res, env, hist, &end_input, &end_symparse);

        // process activation result
        if ( *ret_comp == 00 )
          *ret_comp = get_processed_token (res_proc, env, hist, &end_input, &end_symparse);
      }
    }

    // if a token was recognized, reset parsers to restart recognition
    if ( res_msg == BIN_RECOGNITION_TOKEN_SUCC ) {
      // send a message to components
      m_recogs.start_activate_all ();
      while (m_recogs.activate_all ( get_from BIN_SYNC_RECOGNITION, env, hist, 00, FALSE )) {}
    }

  } while (m_recogs.use_next_type () && ! end_symparse);

  return end_input;
}



/*
 * Get input (char by char) from a file handler, analise it,
 * create a list with the respective components.
 * data_source - from where the data come
 * ct - table with character classes
 * env - sbase enviroment
 * hist - the history
 * 
 * return:
 * the recognized component or 00 if none
 */
class_component *class_translator::
translate_input (FILE* data_source,
                 class_chclass_table* ct,
                 class_enviroment* env,
                 class_sbase_history* hist)
{
  // where to store the symbol read
  symbol_info sym;
  sym.ct = ct;
  sym.obj.set (1);

  // for each character read from input
  BOOL end_input = FALSE;
  class_component* ret_comp = 00; // component to be returned when token found
  while (! end_input ) {  // end if read 0, or EOF

    // read symbol
    sym.obj.import (data_source);
    char ch = sym.obj.get ();

    // send the symbol to parsers
    if ( ch != 0 )
      end_input = parse_symbol (&sym, env, hist, &ret_comp);
    else
      end_input = TRUE;
  }

  // the new component
  return ret_comp;
}
